﻿# include <Siv3D.hpp> // Siv3D v0.6.14

double timeScale = 1.0;

double groundY = 550;

double dt() { return Max(timeScale * Scene::DeltaTime(), 0.0); }

Vec2 polar(double angle) { return Vec2{ Cos(angle * 1_deg),Sin(angle * 1_deg) }; }

double getAngle(Vec2 v) { return (v.getAngle() + 90_deg) / 1_deg; }

double Variance(double var) { return Random(-Abs(var), Abs(var)); }

//移動に関する構造体
struct Transform
{
	Vec2 pos{ 0,0 };

	Vec2 vel{ 0,0 };

	void addVel(double speed,double angle)
	{
		vel += speed * polar(angle);
	}

	void setVel(double speed, double angle)
	{
		vel = speed * polar(angle);
	}

	void update()
	{
		pos += vel * dt();
	}
};

//物体の基底の構造体
struct Object
{
	Transform transform;

	Circle hitbox;

	void setHitboxR(double radius)
	{
		hitbox = Circle{ radius };
	}

	void updateTransform()
	{
		transform.update();
		hitbox.setCenter(transform.pos);
	}

	void drawHitbox(ColorF color)
	{
		hitbox.draw(color);
	}
};

struct Player;

struct PlayerController
{
	Player* owner;

	virtual void update() {};
};
//プレイヤーのひな型
struct Player :Object
{
	//濡れ度合
	int32 wet = 0.0;

	Texture looks;

	Texture runLooks;

	static double scale;

	//移動を制御
	PlayerController* controller = nullptr;

	//向いてる方向
	bool right = true;

	~Player()
	{
		if (controller != nullptr)delete controller;
	}

	void setController(PlayerController* cont)
	{
		if (controller != nullptr)delete controller;
		controller = cont;
		controller->owner = this;
	}

	Vec2 getFootPos()
	{
		return transform.pos + Vec2{ 0,looks.height() * scale / 2 };
	}

	void setFootPos(Vec2 p)
	{
		transform.pos = p + Vec2{ 0,-looks.height() * scale / 2 };
	}

	void jump(double speed, bool touchGroundCheck = true)
	{
		if ((getFootPos().y == groundY) or (not touchGroundCheck))
		{
			transform.vel.y = -speed;
		}
	}

	void update()
	{
		if (getFootPos().y < groundY)
		{
			transform.addVel(1300 * dt(), 90);
		}
		else {
			transform.vel.y = 0;
			setFootPos({ transform.pos.x, groundY });
		}

		if (controller != nullptr)controller->update();

		if (transform.pos.x <= 0)
		{
			transform.vel.x = Max(0.0, transform.vel.x);
		}
		if (Scene::Width() <= transform.pos.x)
		{
			transform.vel.x = Min(0.0, transform.vel.x);
		}

		if (transform.vel.x > 0)
		{
			right = true;
		}
		else if (transform.vel.x < 0)
		{
			right = false;
		}

		updateTransform();
	}

	void draw()
	{
		if (Abs(transform.vel.x) > 0)
		{
			runLooks.mirrored(right).scaled(scale).drawAt(transform.pos);
		}
		else {
			looks.mirrored(right).scaled(scale).drawAt(transform.pos);
		}
	}
};

double Player::scale = 0.85;

//キー入力に応じてプレイヤーを動かす構造体
struct KeyController :PlayerController
{
	Input upKey=KeyUp;
	Input rightKey = KeyRight;
	Input leftKey = KeyLeft;
	double upKeyDur = 0;
	void update() override
	{
		//ジャンプ
		if (upKeyDur > 0)
		{
			upKeyDur += dt();

			if (upKeyDur > 0.11)
			{
				//空中でもう一度ジャンプさせることで強ジャンプ
				owner->jump(630, false);
				upKeyDur = 0;
			}
			else if (upKey.up())
			{
				upKeyDur = 0;
			}
		}
		else if (owner->getFootPos().y == groundY and upKey.down())
		{
			owner->jump(400);
			upKeyDur += dt();
		}
	
		owner->transform.vel.x = 0;

		if (rightKey.pressed())
		{
			owner->transform.vel.x += 300;
		}
		if (leftKey.pressed())
		{
			owner->transform.vel.x -= 300;
		}
	}
};

//コンピュータプレイヤーを動かす構造体
struct RandomCpController :PlayerController
{
	double timer=0.0;

	double jumpTimer = 0.0;

	double stopTimer = 0.0;

	void update() override
	{
		timer += dt();
		jumpTimer += dt();
		stopTimer += dt();

		if (timer >= 0.4)
		{
			//右に行く確率　右にいたら左に行きやすい　逆もしかり
			double p = 0.8 - (0.8 * 2 - 1) * owner->transform.pos.x / Scene::Width();
			
			bool right = RandomBool(p);

			owner->transform.vel.x = right ? 300 : -300;

			timer = Random(0.35);
		}
		//ジャンプ
		if (jumpTimer >= 5.0)
		{
			owner->jump(RandomBool(0.6) ? 700 : 400);
			jumpTimer = Random(3.2);
		}

		if (stopTimer >= 3.0)
		{
			if (RandomBool())
			{
				owner->transform.vel.x = 0;
				//0.2秒止まったあとに動き出す
				timer = Random(0.13);
			}
			stopTimer = Random(1.5);
		}
	}
};

struct Umbrella :public Object
{
	//持ち主
	Player* holder = nullptr;

	Texture looks{ U"☂️"_emoji };

	void update()
	{
		transform.pos = holder->transform.pos + Vec2{ holder->right ? 15 : -15,-50 };
		updateTransform();
	}

	void draw()
	{
		looks.drawAt(transform.pos);
	}
};

//コンピュータプレイヤーを動かす構造体
struct CpController :PlayerController
{
	Player* other;//相手

	Umbrella* umbrella;

	double timer = 0.0;

	double jumpTimer = 0.0;

	double stopTimer = 0.0;

	void update() override
	{
		jumpTimer += dt();
		timer += dt();
		//
		double dx = owner->transform.pos.x - other->transform.pos.x;
		//ジャンプ
		if (jumpTimer >= 5.0)
		{
			//プレイヤーが近いほどつぎのジャンプまでが早くなる->頻繁にジャンプする
			owner->jump(RandomBool(0.7) ? 700 : 400);
			jumpTimer = Random(1.5 + (1 - Abs(dx) / Scene::Width()));
		}

		//自分が傘を持ってるか
		if (umbrella->holder == owner)
		{
			//左右
			if (timer >= 0.4)
			{
				//右に行く確率　右にいたら左に行きやすい　逆もしかり
				double p = 1 - owner->transform.pos.x / Scene::Width();
				//プレイヤーから離れようとする
				p += 0.4 * dx / (Abs(dx) + 1);
				//プレイヤー

				bool right = RandomBool(p);

				owner->transform.vel.x = right ? 300 : -300;

				timer = Random(0.3);
			}
		}
		else //傘を持ってない場合
		{
			//左右　プレイヤーを追いかけろ！
			if (timer >= 0.4)
			{
				//右に行く確率　右にいたら左に行きやすい　逆もしかり
				double p = 1 - owner->transform.pos.x / Scene::Width();
				//プレイヤーへ向かいやすくする
				p -= 0.4 * dx / (Abs(dx) + 1);

				bool right = RandomBool(p);

				owner->transform.vel.x = right ? 300 : -300;

				timer = Random(0.3);
			}
		}
	}
};

//雨のひな型
struct Rain
{
	Array<Object> rainDrops;
	//雨粒を降らせる角度の平均
	double angle = 90;
	//角度のばらつき
	double angleVariance = 8;
	//雨粒の速度の平均
	double speed = 150;
	//速度のばらつき
	double speedVariance = 50;
	//雨粒生成の時間間隔
	double geneTimeInterval = 0.4;
	//生成タイマー
	double geneTimer = 0;
	//ゲームのモード　相合傘モードか略奪モード
	String gameMode;

	void changeWheather(double s, double ang,double interval)
	{
		speed = s;
		speedVariance = s / 3;
		angle = ang;
		geneTimeInterval = interval;
	}

	//雨の主処理
	void update()
	{
		geneTimer += dt();
		//タイマーが生成インターバルをこえたら
		while(geneTimer > geneTimeInterval)
		{
			Object rainDrop{
				.transform = Transform{ Vec2{ Random(0,Scene::Width()),Random(-50,-20) } },
				.hitbox = Circle{5}
			};
			rainDrop.transform.setVel(speed + Variance(speedVariance), angle + Variance(angleVariance));

			rainDrops << rainDrop;
			//タイマーを0にしてリセット
			geneTimer -= geneTimeInterval;
		}
		//雨粒を動かす
		for (int32 i = 0; i < rainDrops.size(); i++)
		{
			rainDrops[i].updateTransform();
			if (rainDrops[i].transform.pos.y >= groundY)
			{
				rainDrops.remove_at(i);
				i--;
			}
		}
	}

	void draw()
	{
		//雨粒を表示
		for (int32 i = 0; i < rainDrops.size(); i++)
		{
			rainDrops[i].drawHitbox(ColorF{ 0.3,0.4,0.7 });
		}
	}

	bool intersects(Circle shape)
	{
		bool hit = false;
		for (int32 i = 0; i < rainDrops.size(); i++)
		{
			//当たってたら
			if (rainDrops[i].hitbox.intersects(shape))
			{
				rainDrops.remove_at(i);//rain[i]を削除
				i--;//一個下がる
				hit = true;
			};
		}
		return hit;
	}
};

struct mySceneManager;

//シーンのひな形
struct myScene
{
	mySceneManager* owner = nullptr;

	virtual void update() {};
};

struct mySceneManager
{
	myScene* nowScene = nullptr;

	~mySceneManager() {
		if (nowScene != nullptr)delete nowScene;
	}

	void ChangeScene(myScene* scene)
	{
		scene->owner = this;
		if (nowScene != nullptr)delete nowScene;
		nowScene = scene;
	}

	void update()
	{
		nowScene->update();
	}
};

//ゲーム画面の構造体
struct Game :myScene
{
	//雨
	Rain rain;
	//経過時間
	double timer = 0;
	//1回雨粒にあたるとプレイヤーの濡れ度合いが2.5上がる
	//プレイヤー１
	Player player1;
	//プレイヤー２
	Player player2;
	//傘　最初はプレイヤー2に持たせる
	Umbrella umbrella;

	String gameMode;
	//天候が変わる時間とタイマー
	double wheatherTime=9;
	double wheatherTimer = 0;
	//大雨警報用のフォント
	Font alertRainfall{ 35,Typeface::Heavy };
	double alertTimer = 0;

	//クリア条件の時間 だいだい6回くらい天候が変わったら終了
	double timeForGameClear = 9 * 7;

	//略奪タイマー
	double stealTimer=0;

	Font resultFont{ 45 ,Typeface::Bold };

	Audio bgm{ Resource(U"悪天候.mid") };

	void update() override
	{
		timer += dt();

		rain.update();

		player1.update();

		player2.update();

		umbrella.update();

		if (alertTimer > 0)
		{
			alertTimer += dt();

			//9回点滅したら警報を止める
			if (alertTimer >= 5.4)
			{
				alertTimer = 0;
			}
		}

		wheatherTimer += dt();
		if (wheatherTime <= wheatherTimer)
		{
			double speed = Random(120, 300);
			double t = Random(0.3) - 0.06;
			rain.changeWheather(speed, 90 + Variance(10), Max(0.009, t));

			if (Max(0.009, t) < 0.01)
			{
				alertTimer += dt();
				wheatherTimer = 0;
			}
			else
			{
				wheatherTimer = Random(wheatherTime / 4) + wheatherTime / 2;
			}
		}

		if (timer > timeForGameClear - 5)
		{
			rain.changeWheather(0, 0, Math::Inf);//雨が止む
			wheatherTimer = 0;//wheatherTimerを止める
		}

		rain.intersects(umbrella.hitbox);

		//傘を持っていないプレイヤー
		Player* notHavingUmbrella = &player1;
		if (umbrella.holder == notHavingUmbrella)notHavingUmbrella = &player2;

		//傘を持っていないやつの当たり判定
		if (rain.intersects(notHavingUmbrella->hitbox))
		{
			notHavingUmbrella->wet++;
		}

		//傘を持っていないやつが傘の中に入っていたら
		if (notHavingUmbrella->hitbox.intersects(umbrella.hitbox))
		{
			stealTimer += dt();
		}
		else
		{
			stealTimer = 0;
		}
		//0.3秒以上たったら
		if (stealTimer > 0.3)
		{
			if(gameMode==U"略奪モード")umbrella.holder = notHavingUmbrella;
			stealTimer = 0;
		}

		player1.draw();

		player2.draw();

		rain.draw();

		//警報を描く
		if (alertTimer > 0)
		{
			alertRainfall(U"大雨警報発令中").drawAt(
													Scene::Center() + Vec2{ 0,-200 },
													ColorF{ Palette::Red,Fmod(alertTimer,0.6) / 0.6 }
													);
		}


		umbrella.draw();

		if (timer > timeForGameClear)
		{
			if (gameMode == U"相合傘モード")
			{
				drawResult(Scene::CenterF() + Vec2 { 0, -200 }, U"Player1", Palette::Aqua, player1.wet);
			}
			else
			{
				drawResult(Scene::CenterF() + Vec2{ 0,-200 }, U"Player1（青い服の人）", Palette::Aqua,player1.wet);
				drawResult(Scene::CenterF() + Vec2{ 0,0 }, U"Player2（赤い服の人）", Palette::Orange, player2.wet);
			}

			if (SimpleGUI::Button(U"タイトルへ", Scene::Center() + Vec2{ 0,200 }))
			{
				bgm.stop();
				returnTitle();
			}
		}

	}

	void drawResult(Vec2 drawPos, String name, ColorF color, int32 wet)
	{
		String comment;

		ColorF commentColor;

		if (wet >= 20)
		{
			comment = U"びしょぬれ";
			commentColor = Palette::Blue;
		}
		else if (wet >= 12)
		{
			comment = U"靴の中べちゃべちゃ……";
			commentColor = Palette::Darkgreen;
		}
		else if (wet > 0)
		{
			comment = U"水も滴るイイ男！";
			commentColor = Palette::Aqua;
		}
		else
		{
			comment = U"You Are 雨避けマスター!!";
			commentColor = Palette::Red;
		}

		resultFont(U"{}・・・{}滴"_fmt(name, wet)).drawAt(drawPos, color);

		resultFont(U"コメント：{}"_fmt(comment)).drawAt(drawPos + Vec2{ 0,70 }, ColorF{ commentColor });
	}

	void returnTitle();
};

//title画面の構造体
struct Title :myScene
{
	int32 buttonType = 0;

	bool singlePlay = true;

	String gameMode;

	Font titleFont{ 42,Typeface::Bold };

	Font expFont{ 20 };

	void update() override
	{
		titleFont(U"傘忘れちまった").drawAt(Scene::Center() + Vec2{ 0,-200 }, Palette::Aqua);

		expFont(U"雨が止むまで雨粒をよけてください。").drawAt(Scene::Center()+Vec2{0,-100}, Palette::White);

		expFont(U"青い人を←→↑の三つのキーで操作します").drawAt(Scene::Center()+Vec2{0,-50}, Palette::Blue);

		expFont(U"二人プレイの場合、赤い人をA D W の三つのキーで操作できます").drawAt(Scene::Center(), Palette::Red);

		expFont(U"略奪モードでは相手から傘を奪うことができます").drawAt(Scene::Center()+Vec2{0,50}, Palette::White);

		if (buttonType == 0)
		{
			if (SimpleGUI::ButtonAt(U"一人で遊ぶ", Scene::Center() + Vec2{ 0,100 }))
			{
				singlePlay = true;
				buttonType = 1;
			}
			if (SimpleGUI::ButtonAt(U"二人で遊ぶ", Scene::Center() + Vec2{ 0,150 }))
			{
				singlePlay = false;
				buttonType = 1;
			}
			if (SimpleGUI::ButtonAt(U"ゲームを閉じる", Scene::Center() + Vec2{ 0,200 }))
			{
				System::Exit();
			}
		}
		else if (buttonType == 1)
		{
			if (SimpleGUI::ButtonAt(U"相合傘モード", Scene::Center() + Vec2{ 0,100 }))
			{
				gameMode = U"相合傘モード";
				buttonType = 2;
			}
			if (SimpleGUI::ButtonAt(U"略奪モード", Scene::Center() + Vec2{ 0,150 }))
			{
				gameMode = U"略奪モード";
				buttonType = 2;
			}
			if (SimpleGUI::ButtonAt(U"もどる", Scene::Center() + Vec2{ 0,200 }))
			{
				buttonType = 0;
			}
		}
		else if (buttonType == 2)
		{
			Game* scene = new Game;

			scene->gameMode = gameMode;

			double height = Texture{ U"🧍"_emoji }.size().y * Player::scale;
			Player& p1 = scene->player1;
			Player& p2 = scene->player2;
			Umbrella& umb = scene->umbrella;

			p1.looks = Texture{ U"🧍‍♂"_emoji };
			p1.runLooks = Texture{ U"🏃‍♂️"_emoji };
			p1.setFootPos({ 460,groundY });
			p1.setController(new KeyController);
			p1.setHitboxR(10);

			p2.looks = Texture{ U"🧍"_emoji };
			p2.runLooks = Texture{ U"🏃"_emoji };
			p2.setFootPos({ 340,groundY });
			p2.setHitboxR(10);
			if (singlePlay)
			{
				//コントローラを分ける
				if (gameMode == U"略奪モード")
				{
					CpController* controller = new CpController;
					controller->other = &p1;
					controller->umbrella = &umb;
					p2.setController(controller);
				}
				else
				{
					RandomCpController* controller = new RandomCpController;
					p2.setController(controller);
				}				
			}
			else
			{
				KeyController* controller = new KeyController;
				controller->upKey = KeyW;
				controller->leftKey = KeyA;
				controller->rightKey = KeyD;
				p2.setController(controller);
			}
			
			umb.holder = &p2;
			umb.setHitboxR(60);

			scene->bgm.play();

			owner->ChangeScene(scene);
		}
	}
};

void Game::returnTitle()
{
	owner->ChangeScene(new Title);
}

void Main()
{
	// 背景の色を設定する
	Scene::SetBackground(ColorF{ 0.75,0.75,0.75 });

	System::SetTerminationTriggers(UserAction::CloseButtonClicked);

	mySceneManager manager;

	manager.ChangeScene(new Title);

	while (System::Update())
	{
		manager.update();
	}
}
